import warnings
from typing import Any, Callable, Dict, Iterable, List, NamedTuple, Union

from pylspci.device import Device
from pylspci.fields import NameWithID, Slot, hexstring
from pylspci.parsers.base import Parser

UNKNOWN_FIELD_WARNING = (
    'Unsupported device field {!r} with value {!r}\n'
    'Please report this, along with the output of `lspci -mmnnvvvk`, at '
    'https://gitlab.com/Lucidiot/pylspci/issues'
)


class FieldMapping(NamedTuple):
    """
    Helper class to map verbose output field names such as ``SVendor`` to
    :class:`Device` fields such as ``subsytem_vendor``.
    """

    field_name: str
    """
    Field name on the :class:`Device` named tuple.
    """

    field_type: Callable[[str], Any]
    """
    Field type; a callable to use to parse the string value.
    """

    many: bool = False
    """
    Whether or not to use a List, if this field can be repeated multiple times
    in the lspci output.
    """


class VerboseParser(Parser):
    """
    A parser for lspci -vvvmmk
    """

    default_lspci_args = {
        'verbose': True,
        'kernel_drivers': True,
    }

    # Maps lspci output fields to Device fields with a type
    _field_mapping = {
        'Slot': FieldMapping(field_name='slot', field_type=Slot),
        'Class': FieldMapping(field_name='cls', field_type=NameWithID),
        'Vendor': FieldMapping(field_name='vendor', field_type=NameWithID),
        'Device': FieldMapping(field_name='device', field_type=NameWithID),
        'SVendor': FieldMapping(
            field_name='subsystem_vendor',
            field_type=NameWithID,
        ),
        'SDevice': FieldMapping(
            field_name='subsystem_device',
            field_type=NameWithID,
        ),
        'Rev': FieldMapping(field_name='revision', field_type=hexstring),
        'ProgIf': FieldMapping(field_name='progif', field_type=hexstring),
        'Driver': FieldMapping(field_name='driver', field_type=str),
        'Module': FieldMapping(
            field_name='kernel_modules',
            field_type=str,
            many=True,
        ),
        'NUMANode': FieldMapping(field_name='numa_node', field_type=int),
        'IOMMUGroup': FieldMapping(field_name='iommu_group', field_type=int),
        'PhySlot': FieldMapping(field_name='physical_slot', field_type=str),
    }

    def _parse_device(self, device_data: Union[str, Iterable[str]]) -> Device:
        devdict: Dict[str, Any] = {}
        if isinstance(device_data, str):
            device_data = device_data.splitlines()

        for line in device_data:
            key, _, value = map(str.strip, line.partition(':'))
            if key not in self._field_mapping:
                warnings.warn(
                    UNKNOWN_FIELD_WARNING.format(key, value),
                    UserWarning,
                )
                continue
            field = self._field_mapping[key]
            if field.many:
                devdict.setdefault(field.field_name, []) \
                       .append(field.field_type(value))
            else:
                devdict[field.field_name] = field.field_type(value)

        return Device(**devdict)

    def parse(
            self,
            data: Union[str, Iterable[str], Iterable[Iterable[str]]],
            ) -> List[Device]:
        """
        Parse an lspci -vvvmm[nnk] output, either as a single string holding
        multiple devices separated by two newlines,
        or as a list of multiline strings holding one device each.

        :param data: A string holding multiple devices,
           a list of strings, one for each device,
           or a list of lists of strings, one list for each device, with
           each list holding each part of the device output.
        :type data: str or Iterable[str] or Iterable[Iterable[str]]
        :return: A list of parsed devices.
        :rtype: List[Device]
        """
        if isinstance(data, str):
            data = data.split('\n\n')
        result: List[Device] = []
        for line in data:
            if isinstance(line, str):
                line = str.strip(line)
            if not line:  # Ignore empty strings and lists
                continue
            result.append(self._parse_device(line))
        return result



from typing import Dict, List, NamedTuple, Optional, Union

from pylspci.fields import NameWithID, NameWithIDDict, Slot, SlotDict

DeviceDict = Dict[str, Union[
    int,
    str,
    SlotDict,
    NameWithIDDict,
    List[str],
    None,
]]


class Device(NamedTuple):
    """
    Describes a device returned by lspci.
    """

    slot: Slot
    """
    The device's slot (domain, bus, number and function).
    """

    cls: NameWithID
    """
    The device's class, with a name and/or an ID.
    """

    vendor: NameWithID
    """
    The device's vendor, with a name and/or an ID.
    """

    device: NameWithID
    """
    The device's name and/or ID.
    """

    subsystem_vendor: Optional[NameWithID] = None
    """
    The device's subsystem vendor, if found, with a name and/or an ID.
    """

    subsystem_device: Optional[NameWithID] = None
    """
    The device's subsystem name and/or ID, if found.
    """

    revision: Optional[int] = None
    """
    The device's revision number.
    """

    progif: Optional[int] = None
    """
    The device's programming interface number.
    """

    driver: Optional[str] = None
    """
    The device's driver (Linux only).
    """

    kernel_modules: List[str] = []
    """
    One or more kernel modules that can handle this device (Linux only).
    """

    numa_node: Optional[int] = None
    """
    NUMA node this device is connected to (Linux only).
    """

    iommu_group: Optional[int] = None
    """
    IOMMU group that this device is part of (optional, Linux only).
    """

    physical_slot: Optional[str] = None
    """
    The device's physical slot number (Linux only).
    """

    def as_dict(self) -> DeviceDict:
        """
        Serialize this device as a JSON-serializable `dict`.
        """
        return {
            "slot": self.slot.as_dict(),
            "cls": self.cls.as_dict(),
            "vendor": self.vendor.as_dict(),
            "device": self.device.as_dict(),
            "subsystem_vendor": (
                self.subsystem_vendor.as_dict()
                if self.subsystem_vendor
                else None
            ),
            "subsystem_device": (
                self.subsystem_device.as_dict()
                if self.subsystem_device
                else None
            ),
            "revision": self.revision,
            "progif": self.progif,
            "driver": self.driver,
            "kernel_modules": self.kernel_modules,
            "numa_node": self.numa_node,
            "iommu_group": self.iommu_group,
            "physical_slot": self.physical_slot,
        }



import subprocess
from enum import Enum
from pathlib import Path
from typing import (
    Any, Iterator, List, Mapping, MutableMapping, Optional, Union
)

from pylspci.device import Device
from pylspci.fields import PCIAccessParameter
from pylspci.filters import DeviceFilter, SlotFilter
from pylspci.parsers.base import Parser

OptionalPath = Optional[Union[str, Path]]


class IDResolveOption(Enum):
    """
    ``lspci`` device, vendor, class names outputting options.
    """

    NameOnly = ''
    """
    Only output the names.
    """

    IDOnly = '-n'
    """
    Only output the hexadecimal IDs.
    This is the only option that does not require a ``pciids`` file.
    """

    Both = '-nn'
    """
    Output both the names and hexadecimal IDs, in the format ``Name [ID]``.
    """


def lspci(
        pciids: OptionalPath = None,
        pcimap: OptionalPath = None,
        access_method: Optional[str] = None,
        pcilib_params: Mapping[str, Any] = {},
        file: OptionalPath = None,
        verbose: bool = False,
        kernel_drivers: bool = False,
        bridge_paths: bool = False,
        hide_single_domain: bool = True,
        id_resolve_option: IDResolveOption = IDResolveOption.Both,
        slot_filter: Optional[Union[SlotFilter, str]] = None,
        device_filter: Optional[Union[DeviceFilter, str]] = None,
        ) -> str:
    """
    Call the ``lspci`` command with various parameters.

    :param pciids: An optional path to a ``pciids`` file,
       to convert hexadecimal class, vendor or device IDs into names.
    :type pciids: str or Path or None
    :param pcimap: An optional path to a ``pcimap`` file,
       linking Linux kernel modules and their supported PCI IDs.
    :type pcimap: str or Path or None
    :param access_method: The access method to use to find devices.
       Set this to ``help`` to list the available access methods in a
       human-readable format. For the machine-readable format, see
       :func:`list_access_methods`.
    :type access_method: str or None
    :param pcilib_params: Parameters passed to pcilib's access methods.
       To list the available parameters with their description and default
       values, see :func:`list_pcilib_params`.
    :type pcilib_params: Mapping[str, Any] or None
    :param file: An hexadecimal dump from ``lspci -x`` to load data from,
       instead of accessing real hardware.
    :type file: str or Path or None
    :param bool verbose: Increase verbosity.
       This radically changes the output format.
    :param bool kernel_drivers: Also include kernel modules and drivers
       in the output. Only has effect with the verbose output.
    :param bool bridge_paths: Add PCI bridge paths to slot numbers.
    :param bool hide_single_domain: If there is a single PCI domain on this
       machine and it is numbered ``0000``, hide it from the slot numbers.
    :param id_resolve_option: Device, vendor or class ID outputting mode.
       See the :class:`IDResolveOption` docs for more details.
    :type id_resolve_option: IDResolveOption
    :param slot_filter: Filter devices by their slot
      (domain, bus, device, function)
    :type slot_filter: SlotFilter or str or None
    :param device_filter: Filter devices by their vendor, device or class ID
    :type device_filter: DeviceFilter or str or None
    :return: Any output from the ``lspci`` command.
    :rtype: str
    :raises subprocess.CalledProcessError:
       ``lspci`` returned a non-zero error code.
    """
    args: List[str] = ['lspci', '-mm']
    if verbose:
        args.append('-vvv')
    if kernel_drivers:
        args.append('-k')
    if bridge_paths:
        args.append('-PP')
    if not hide_single_domain:
        args.append('-D')
    if access_method:
        args.append('-A{}'.format(access_method))
    if slot_filter:
        if isinstance(slot_filter, str):
            slot_filter = SlotFilter.parse(slot_filter)
        args.append('-s')
        args.append(str(slot_filter))
    if device_filter:
        if isinstance(device_filter, str):
            device_filter = DeviceFilter.parse(device_filter)
        args.append('-d')
        args.append(str(device_filter))
    if id_resolve_option != IDResolveOption.NameOnly:
        args.append(id_resolve_option.value)

    if pciids:
        args.append('-i')
        if not isinstance(pciids, Path):
            pciids = Path(pciids)
        assert pciids.is_file(), 'ID database file not found'
        args.append(str(pciids.absolute()))

    if pcimap:
        args.append('-p')
        if not isinstance(pcimap, Path):
            pcimap = Path(pcimap)
        assert pcimap.is_file(), 'Kernel module mapping file not found'
        args.append(str(pcimap.absolute()))

    if file:
        args.append('-F')
        if not isinstance(file, Path):
            file = Path(file)
        assert file.is_file(), 'Hex dump file not found'
        args.append(str(file.absolute()))

    for key, value in pcilib_params.items():
        args.append('-O{}={}'.format(key, value))

    return subprocess.check_output(
        args,
        universal_newlines=True,
    )


def list_access_methods() -> List[str]:
    """
    Calls ``lspci(access_method='help')`` to list the PCI access methods
    the underlying ``pcilib`` provides and parses the human-readable list into
    a machine-readable list.

    :returns: A list of access methods.
    :rtype: List[str]
    :raises subprocess.CalledProcessError:
       ``lspci`` returned a non-zero error code.
    """
    return list(filter(
        lambda line: line and 'Known PCI access methods' not in line,
        map(str.strip, lspci(access_method='help').splitlines()),
    ))


def list_pcilib_params_raw() -> List[str]:
    """
    Calls ``lspci -Ohelp`` to list the PCI access parameters the underlying
    ``pcilib`` provides.

    :returns: A list of available PCI access parameters.
    :rtype: List[str]
    :raises subprocess.CalledProcessError:
       ``lspci`` returned a non-zero error code.
    """
    return list(filter(
        lambda line: line and 'Known PCI access parameters' not in line,
        map(str.strip, subprocess.check_output(
            ['lspci', '-Ohelp'],
            universal_newlines=True,
        ).splitlines())
    ))


def list_pcilib_params() -> List[PCIAccessParameter]:
    """
    Calls ``lspci -Ohelp`` to list the PCI access parameters the underlying
    ``pcilib`` provides and parse the human-readable list into
    a machine-readable list.

    :returns: A list of available PCI access parameters.
    :rtype: List[PCIAccessParameter]
    :raises subprocess.CalledProcessError:
       ``lspci`` returned a non-zero error code.
    """
    return list(map(PCIAccessParameter, list_pcilib_params_raw()))


class CommandBuilder(object):
    """
    Helper class to build a lspci call using a Builder pattern.

    Iterating over the builder will result in the command being called,
    and will return strings, devices or pcilib parameters, one at a time,
    depending on the parsing settings.
    """

    _list_access_methods: bool = False
    _list_pcilib_params: bool = False
    _list_pcilib_params_raw: bool = False
    _params: MutableMapping[str, Any] = {}
    _parser: Optional[Parser] = None

    def __init__(self, **kwargs: Any):
        self._params = kwargs

    def __iter__(self) -> Iterator[Union[str, Device, PCIAccessParameter]]:
        result: Union[str, List[str], List[Device], List[PCIAccessParameter]]
        if self._list_access_methods:
            result = list_access_methods()
        elif self._list_pcilib_params:
            if self._list_pcilib_params_raw:
                result = list_pcilib_params_raw()
            else:
                result = list_pcilib_params()
        elif self._parser:
            result = self._parser.parse(lspci(**self._params))
        else:
            result = lspci(**self._params)

        if isinstance(result, str):
            return iter([result, ])
        return iter(result)

    def use_pciids(self,
                   path: OptionalPath,
                   check: bool = True) -> 'CommandBuilder':
        """
        Use a PCI IDs file from a given path.

        :param path: A string or path-like object pointing to the PCI IDs file
           to use. Set to None to use the default files from lspci.
        :type path: str or Path or None
        :param bool check: Whether to check for the file's existence
           immediately, or delay that to the lspci invocation.
        :returns: The current CommandBuilder instance.
        :rtype: CommandBuilder
        """
        if path:
            if not isinstance(path, Path):
                path = Path(path)
            if check:
                assert path.is_file(), 'ID database file not found'
        self._params['pciids'] = path
        return self

    def use_pcimap(self,
                   path: OptionalPath,
                   check: bool = True) -> 'CommandBuilder':
        """
        Use a kernel module mapping file from a given path.

        :param path: A string or path-like object pointing to the mapping file
           to use. Set to None to use the default files from lspci.
        :type path: str or Path or None
        :param bool check: Whether to check for the file's existence
           immediately, or delay that to the lspci invocation.
        :returns: The current CommandBuilder instance.
        :rtype: CommandBuilder
        """
        if path:
            if not isinstance(path, Path):
                path = Path(path)
            if check:
                assert path.is_file(), 'Kernel module mapping file not found'
        self._params['pcimap'] = path
        return self

    def use_access_method(self, method: Optional[str]) -> 'CommandBuilder':
        """
        Use a specific access method to list all devices.

        :param method: Name of the access method to use. Set to None to use
           lspci's default method.
        :type method: str or None
        :returns: The current CommandBuilder instance.
        :rtype: CommandBuilder
        """
        self._params['access_method'] = method
        return self

    def list_access_methods(self, value: bool = True) -> 'CommandBuilder':
        """
        List the pcilib access methods instead of listing devices.

        :param value: Whether or not to list the access methods.
        :type value: bool
        :returns: The current CommandBuilder instance.
        :rtype: CommandBuilder
        """
        self._list_access_methods = value
        if value:
            self._list_pcilib_params = False
        return self

    def list_pcilib_params(self,
                           value: bool = True,
                           raw: bool = False) -> 'CommandBuilder':
        """
        List the pcilib parameters instead of listing devices.

        :param value: Whether or not to list the pcilib parameters.
        :type value: bool
        :param raw: When listing the pcilib parameters, whether to return the
           raw strings or parse them into PCIAccessParameter instances.
        :type raw: bool
        :returns: The current CommandBuilder instance.
        :rtype: CommandBuilder
        """
        self._list_pcilib_params = value
        self._list_pcilib_params_raw = raw
        if value:
            self._list_access_methods = False
        return self

    def with_pcilib_params(self,
                           *args: Mapping[str, Any],
                           **kwargs: Any) -> 'CommandBuilder':
        """
        Override some pcilib parameters. When given a dict, will rewrite the
        parameters with the new dict. When given keyword arguments, will update
        the existing parameters. Pass ``None`` or ``{}`` to reset all of the
        parameters.

        :returns: The current CommandBuilder instance.
        :rtype: CommandBuilder
        """
        if len(args) > 0:
            assert len(args) <= 1, 'Only one positional argument is allowed'
            assert not kwargs, 'Use either a dict or keyword arguments'
            self._params['pcilib_params'] = args[0]
            return self
        self._params.setdefault('pcilib_params', {}).update(kwargs)
        return self

    def from_file(self,
                  path: OptionalPath,
                  check: bool = True) -> 'CommandBuilder':
        """
        Use a hexadecimal dump from a previous run of lspci instead of
        accessing the host's devices directly.

        :param path: A string or path-like object pointing to the hex dump file
           to use. Set to None to not use a dump file.
        :type path: str or Path or None
        :param bool check: Whether to check for the file's existence
           immediately, or delay that to the lspci invocation.
        :returns: The current CommandBuilder instance.
        :rtype: CommandBuilder
        """
        if path:
            if not isinstance(path, Path):
                path = Path(path)
            if check:
                assert path.is_file(), 'Hex dump file not found'
        self._params['file'] = path
        return self

    def verbose(self, value: bool = True) -> 'CommandBuilder':
        """
        Enable verbose mode.

        :param value: Whether or not to use verbose mode.
        :type value: bool
        :returns: The current CommandBuilder instance.
        :rtype: CommandBuilder
        """
        self._params['verbose'] = value
        return self

    def include_kernel_drivers(self, value: bool = True) -> 'CommandBuilder':
        """
        Under Linux, includes the available kernel modules for each device.
        Implies ``.verbose()``.

        :param value: Whether or not to include the available kernel modules
           for each device.
        :type value: bool
        :returns: The current CommandBuilder instance.
        :rtype: CommandBuilder
        """
        self._params['kernel_drivers'] = value
        if value:
            return self.verbose()
        return self

    def include_bridge_paths(self, value: bool = True) -> 'CommandBuilder':
        """
        Include the PCI bridge paths along with the IDs.
        Implies ``.with_ids()``.

        :param value: Whether or not to include the PCI bridge paths.
        :type value: bool
        :returns: The current CommandBuilder instance.
        :rtype: CommandBuilder
        """
        self._params['bridge_paths'] = value
        if value:
            return self.with_ids()
        return self

    def hide_single_domain(self, value: bool = True) -> 'CommandBuilder':
        """
        Hide the domain numbers when there is only one domain, numbered zero.

        :param value: Whether or not to hide the domain numbers for a single
           domain.
        :type value: bool
        :returns: The current CommandBuilder instance.
        :rtype: CommandBuilder
        """
        self._params['hide_single_domain'] = value
        return self

    def with_ids(self, value: bool = True) -> 'CommandBuilder':
        """
        Include PCI device IDs. If disabled, implies ``.with_names()``.

        :param value: Whether or not to include the PCI device IDs.
        :type value: bool
        :returns: The current CommandBuilder instance.
        :rtype: CommandBuilder
        """
        if value:
            if self._params.get('id_resolve_option') == \
                    IDResolveOption.NameOnly:
                self._params['id_resolve_option'] = IDResolveOption.Both
        else:
            self._params['id_resolve_option'] = IDResolveOption.NameOnly
        return self

    def with_names(self, value: bool = True) -> 'CommandBuilder':
        """
        Include PCI device names. If disabled, implies ``.with_ids()``.

        :param value: Whether or not to include the PCI device names.
        :type value: bool
        :returns: The current CommandBuilder instance.
        :rtype: CommandBuilder
        """
        if value:
            if self._params.get('id_resolve_option') == IDResolveOption.IDOnly:
                self._params['id_resolve_option'] = IDResolveOption.Both
        else:
            self._params['id_resolve_option'] = IDResolveOption.IDOnly
        return self

    def slot_filter(self,
                    *args: str,
                    domain: Optional[int] = None,
                    bus: Optional[int] = None,
                    device: Optional[int] = None,
                    function: Optional[int] = None) -> 'CommandBuilder':
        """
        Filter the devices geographically.
        Can be passed a string in lspci's filter syntax, or keyword arguments
        for each portion of the filter.
        See :class:`pylspci.filters.SlotFilter`.

        :returns: The current CommandBuilder instance.
        :rtype: CommandBuilder
        """
        if len(args) > 0:
            assert len(args) <= 1, 'Only one positional argument allowed'
            assert not domain and not bus and not device and not function, \
                'Use either a string value or the domain, bus, device ' \
                'and function keyword arguments'
            self._params['slot_filter'] = SlotFilter.parse(args[0])
        else:
            self._params['slot_filter'] = SlotFilter(
                domain=domain,
                bus=bus,
                device=device,
                function=function,
            )
        return self

    def device_filter(self,
                      *args: str,
                      cls: Optional[int] = None,
                      vendor: Optional[int] = None,
                      device: Optional[int] = None) -> 'CommandBuilder':
        """
        Filter the devices logically.
        Can be passed a string in lspci's filter syntax, or keyword arguments
        for each portion of the filter.
        See :class:`pylspci.filters.DeviceFilter`.

        :returns: The current CommandBuilder instance.
        :rtype: CommandBuilder
        """
        if len(args) > 0:
            assert len(args) <= 1, 'Only one positional argument allowed'
            assert not cls and not vendor and not device, \
                'Use either a string value or the cls, vendor and device ' \
                'keyword arguments'
            self._params['device_filter'] = DeviceFilter.parse(args[0])
        else:
            self._params['device_filter'] = \
                DeviceFilter(cls=cls, vendor=vendor, device=device)
        return self

    def with_parser(self, parser: Optional[Parser] = None) -> 'CommandBuilder':
        """
        Use a pylspci parser to get parsed Device instances instead of strings.

        :param parser: The parser to use. Set to None to disable parsing.
        :type parser: pylspci.parsers.Parser
        :returns: The current CommandBuilder instance.
        :rtype: CommandBuilder
        """
        self._parser = parser
        return self

    def with_default_parser(self) -> 'CommandBuilder':
        """
        Use the default parser compatible with the current set of settings.
        Note that this should be used as one of the last instructions of the
        builder, as the default parser can change if the settings are updated.

        :returns: The current CommandBuilder instance.
        :rtype: CommandBuilder
        """
        if self._params.get('verbose'):
            from pylspci.parsers import VerboseParser
            return self.with_parser(VerboseParser())
        else:
            from pylspci.parsers import SimpleParser
            return self.with_parser(SimpleParser())



import re
from functools import partial
from typing import Any, Dict, Optional, Union

# mypy does not support recursive type definitions
# SlotDict = Dict[str, Union[int, 'SlotDict', None]]
SlotDict = Dict[str, Union[int, Dict[str, Any], None]]
NameWithIDDict = Dict[str, Union[int, str, None]]

hexstring = partial(int, base=16)


class Slot(object):
    """
    Describes a PCI slot identifier, in the format ``[DDDD:]BB:dd.f``,
    where ``D`` is the domain, ``B`` the bus, ``d`` the device
    and ``f`` the function. The first three are hexadecimal numbers, but
    ``f`` is in octal.
    """

    domain: int = 0x0000
    """
    The slot's domain, as a four-digit hexadecimal number.
    When omitted, defaults to ``0x0000``.
    """

    bus: int
    """
    The slot's bus, as a two-digit hexadecimal number.
    """

    device: int
    """
    The slot's device, as a two-digit hexadecimal number, up to `0x1f`.
    """

    function: int
    """
    The slot's function, as a single octal digit.
    """

    parent: Optional["Slot"] = None
    """
    The slot's parent bridge, if present.
    """

    def __init__(self, value: str) -> None:
        parent, _, me = value.rpartition('/')
        if parent:
            self.parent = Slot(parent)

        data = list(map(hexstring, re.split(r'[:\.]', me)))
        if len(data) == 3:
            data.insert(0, self.parent.domain if self.parent else 0)
        self.domain, self.bus, self.device, self.function = data

        if self.device > 0x1f:
            raise ValueError('Device numbers cannot be above 0x1f')
        if self.function > 0x7:
            raise ValueError('Function numbers cannot be above 7')

    def __str__(self) -> str:
        output: str = '{:04x}:{:02x}:{:02x}.{:01x}'.format(
            self.domain, self.bus, self.device, self.function,
        )
        if self.parent:
            return '{!s}/{}'.format(self.parent, output)
        return output

    def __repr__(self) -> str:
        return '{}({!r})'.format(self.__class__.__name__, str(self))

    def as_dict(self) -> SlotDict:
        """
        Serialize this slot as a JSON-serializable `dict`.
        """
        return {
            "domain": self.domain,
            "bus": self.bus,
            "device": self.device,
            "function": self.function,
            "parent": self.parent.as_dict() if self.parent else None,
        }


class NameWithID(object):
    """
    Describes a device, vendor or class with either
    a name, an hexadecimal PCI ID, or both.
    """

    id: Optional[int]
    """
    The PCI ID as a four-digit hexadecimal number.
    """

    name: Optional[str]
    """
    The human-readable name associated with this ID.
    """

    _NAME_ID_REGEX = re.compile(r'^(?P<name>.+)\s\[(?P<id>[0-9a-fA-F]{4})\]$')

    def __init__(self, value: Optional[str]) -> None:
        if value and value.endswith(']'):
            # Holds both an ID and a name
            match = self._NAME_ID_REGEX.match(value)
            if not match:  # Except it doesn't
                self.id = None
                self.name = value
                return
            gd = match.groupdict()
            self.id = hexstring(gd['id'])
            self.name = gd['name']
            return

        try:
            self.id = hexstring(value)
            self.name = None
        except (TypeError, ValueError):
            self.id = None
            self.name = value

    def __str__(self) -> str:
        if self.id and self.name:
            return '{} [{:04x}]'.format(self.name, self.id)
        elif self.name:
            return self.name
        elif self.id:
            return '{:04x}'.format(self.id)
        else:
            return ''

    def __repr__(self) -> str:
        return '{}({!r})'.format(self.__class__.__name__, str(self))

    def as_dict(self) -> NameWithIDDict:
        """
        Serialize this name and ID as a JSON-serializable `dict`.
        """
        return {
            "id": self.id,
            "name": self.name,
        }


class PCIAccessParameter(object):
    """
    A pcilib access method parameter, as parsed from :func:`list_pcilib_params`
    or ``lspci -Ohelp``, that can be modified using the ``pcilib_params``
    argument of :func:`lspci`, or ``lspci -Oname=value`` in the command line.
    """

    name: str
    """
    The parameter's name.
    """

    description: str
    """
    A short description of the parameter's use.
    """

    default: Optional[str]
    """
    An optional default value for the parameter.
    """

    _PARAM_REGEX = re.compile(
        r'^(?P<name>\S+)\s+(?P<description>.+)\s\((?P<default>.*)\)$')

    def __init__(self, value: str) -> None:
        match = self._PARAM_REGEX.match(value)
        if not match:
            raise ValueError(
                'Could not parse {!r} into a parameter'.format(value))
        gd = match.groupdict()
        self.name = gd['name'].strip()
        self.description = gd['description'].strip()
        self.default = gd['default'].strip() or None

    def __str__(self) -> str:
        return '{}\t{} ({})'.format(self.name, self.description, self.default)

    def __repr__(self) -> str:
        return '{!s}({!r})'.format(self.__class__.__name__, str(self))

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, PCIAccessParameter) and \
            (self.name, self.description, self.default) \
            == (other.name, other.description, other.default)

    def as_dict(self) -> Dict[str, Optional[str]]:
        """
        Serialize this PCI access parameter as a JSON-serializable `dict`.
        """
        return {
            "name": self.name,
            "description": self.description,
            "default": self.default,
        }



from abc import ABC, abstractmethod
from typing import Any, Dict, Iterable, List, Union

from pylspci.device import Device


class Parser(ABC):

    default_lspci_args: Dict[str, Any] = {}
    """
    The default arguments that, when sent to :func:`lspci`, should provide the
    best output for this parser.

    See :func:`lspci`'s documentation for a list of available arguments.
    """

    @abstractmethod
    def parse(
            self,
            data: Union[str, Iterable[str], Iterable[Iterable[str]]]
            ) -> List[Device]:
        """
        Parse a string or list of strings as a list of devices.

        :param data: A string holding multiple devices,
           a list of strings, one for each device,
           or a list of lists of strings, one list for each device, with
           each list holding each part of the device output.
        :type data: str or Iterable[str] or Iterable[Iterable[str]]
        :returns: A list of parsed devices.
        :rtype: List[Device]
        """

    def run(self, **kwargs: Any) -> List[Device]:
        """
        Run the lspci command with the given arguments, defaulting to the
        parser's default arguments, and parse the result.

        :param \\**kwargs: Optional arguments to override the parser's default
           arguments. See :func:`lspci`'s documentation for a list of
           available arguments.
        :type \\**kwargs: Any
        :returns: A list of parsed devices.
        :rtype: List[Device]
        """
        from pylspci.command import lspci
        lspci_kwargs = self.default_lspci_args.copy()
        lspci_kwargs.update(kwargs)
        return self.parse(lspci(**lspci_kwargs))



#!/usr/bin/env python3
import argparse
import json
from pathlib import Path
from typing import Any, Dict, Optional

from pylspci.command import CommandBuilder, IDResolveOption
from pylspci.filters import DeviceFilter, SlotFilter


def get_parser() -> argparse.ArgumentParser:
    parser: argparse.ArgumentParser = argparse.ArgumentParser(
        description='Python wrapper for lspci',
    )
    parser.add_argument(
        '-i', '--pci-ids',
        help='Path to an alternate file to use as the PCI ID list.',
        required=False,
        type=Path,
        dest='pciids',
    )
    parser.add_argument(
        '-p', '--pci-map',
        help='Path to an alternate file to use as the '
             'kernel module mapping file.',
        required=False,
        type=Path,
        dest='pcimap',
    )

    filter_options = parser.add_argument_group(title='Filters')
    filter_options.add_argument(
        '-s',
        help='Filter devices by their slots. '
             'Any value can be omitted or set to * to disable filtering.',
        type=SlotFilter.parse,
        dest='slot_filter',
        metavar='[[domain:]bus:][device][.function]',
    )
    filter_options.add_argument(
        '-d',
        help='Filter devices by their type. '
             'Any value can be omitted or set to * to disable filtering.',
        type=DeviceFilter.parse,
        dest='device_filter',
        metavar='[vendor]:[device][:class]',
    )

    output_options = parser.add_argument_group(title='Output options')
    output_options.add_argument(
        '-v', '--verbose',
        help='Display more details about devices.',
        default=False,
        action='store_true',
        dest='verbose',
    )
    output_options.add_argument(
        '-k', '--kernel-modules',
        help='On Linux kernels above 2.6, include kernel drivers handling '
             'each device and kernel modules able to handle them. Implies -v.',
        default=False,
        action='store_true',
        dest='kernel_drivers',
    )
    output_options.add_argument(
        '-P', '-PP', '--bridge-paths',
        help='Include PCI bridge paths along with device IDs.',
        default=False,
        action='store_true',
        dest='bridge_paths',
    )

    output_modes = output_options.add_mutually_exclusive_group()
    output_modes.add_argument(
        '--json',
        help='Parse the lspci output and return JSON data.',
        action='store_true',
        default=True,
        dest='json',
    )
    output_modes.add_argument(
        '--raw',
        help="Return lspci's output directly, without parsing.",
        action='store_false',
        default=True,
        dest='json',
    )

    id_resolve_option = output_options.add_mutually_exclusive_group()
    id_resolve_option.add_argument(
        '--name-only',
        help='Only include device names. This is the default.',
        action='store_const',
        default=IDResolveOption.NameOnly,
        const=IDResolveOption.NameOnly,
        dest='id_resolve_option',
    )
    id_resolve_option.add_argument(
        '-n', '--id-only',
        help='Only include device IDs, without looking for names '
             'in the PCI ID file.',
        action='store_const',
        const=IDResolveOption.IDOnly,
        default=IDResolveOption.NameOnly,
        dest='id_resolve_option',
    )
    id_resolve_option.add_argument(
        '-nn', '--name-with-id',
        help='Include both device IDs and names.',
        action='store_const',
        const=IDResolveOption.Both,
        default=IDResolveOption.NameOnly,
        dest='id_resolve_option',
    )

    access_options = parser.add_argument_group(title='PCI access options')
    access_options.add_argument(
        '-O', '--option',
        help='Set PCI library access parameters. '
             'Use -O help to get a list of available parameters '
             'with their descriptions and default values.',
        action='append',
        dest='pcilib_params',
        metavar='KEY=VALUE',
    )
    access_exclusive = access_options.add_mutually_exclusive_group()
    access_exclusive.add_argument(
        '-A', '--access-method',
        help='PCI library access method to use. '
             'Use -A help to list available access methods.',
        dest='access_method',
        metavar='METHOD',
    )
    access_exclusive.add_argument(
        '-F', '--file',
        help='Use a hex dump file from a previous run of lspci instead of '
             'accessing real hardware.',
        dest='file',
        metavar='FILE',
    )
    access_exclusive.add_argument(
        '-H1',
        help='Access hardware using Intel configuration mechanism 1. '
             'Alias to -A intel-conf1.',
        action='store_const',
        const='intel-conf1',
        dest='access_method',
    )
    access_exclusive.add_argument(
        '-H2',
        help='Access hardware using Intel configuration mechanism 2. '
             'Alias to -A intel-conf2.',
        action='store_const',
        const='intel-conf2',
        dest='access_method',
    )
    return parser


def main() -> None:
    parser: argparse.ArgumentParser = get_parser()
    args: Dict[str, Any] = vars(parser.parse_args())

    json_output: bool = args.pop('json', True)
    kernel_modules: bool = args.pop('kernel_modules', False)
    access_method: Optional[str] = args.pop('access_method', None)
    pcilib_params = args.pop('pcilib_params', []) or []

    builder: CommandBuilder = CommandBuilder(**args)
    if kernel_modules:
        builder = builder.include_kernel_drivers()

    if access_method:
        if access_method.strip().lower() == 'help':
            builder = builder.list_access_methods()
        else:
            builder = builder.use_access_method(access_method)

    for param in pcilib_params:
        if param.strip().lower() == 'help':
            builder = builder.list_pcilib_params(raw=not json_output)
            break
        if '=' not in param:
            parser.error(
                'Invalid PCI access parameter syntax for {!r}'.format(param))
        key, value = map(str.strip, param.split('=', 2))
        builder = builder.with_pcilib_params(**{key: value})

    if json_output:
        builder = builder.with_default_parser()

    result = list(builder)
    if not json_output:  # Raw mode
        for item in result:
            print(item)
        return

    print(json.dumps([
        item if isinstance(item, str) else item.as_dict()
        for item in result
    ]))


if __name__ == '__main__':
    main()



import argparse
import shlex
from typing import Any, Iterable, List, Union

from cached_property import cached_property

from pylspci.device import Device
from pylspci.fields import NameWithID, Slot, hexstring
from pylspci.parsers.base import Parser


class SimpleParser(Parser):
    """
    A parser for lspci -mm.
    """

    @cached_property
    def _parser(self) -> argparse.ArgumentParser:
        p = argparse.ArgumentParser()
        p.add_argument(
            'slot',
            type=Slot,
        )
        p.add_argument(
            'cls',
            type=NameWithID,
        )
        p.add_argument(
            'vendor',
            type=NameWithID,
        )
        p.add_argument(
            'device',
            type=NameWithID,
        )
        p.add_argument(
            'subsystem_vendor',
            type=NameWithID,
        )
        p.add_argument(
            'subsystem_device',
            type=NameWithID,
        )
        p.add_argument(
            '-r',
            type=hexstring,
            nargs='?',
            dest='revision',
        )
        p.add_argument(
            '-p',
            type=hexstring,
            nargs='?',
            dest='progif',
        )
        return p

    def parse(
            self,
            data: Union[str, Iterable[str], Iterable[Iterable[str]]],
            ) -> List[Device]:
        """
        Parse a multiline string or a list of single-line strings
        from lspci -mm into devices.

        :param data: A string holding multiple devices,
           a list of strings, one for each device,
           or a list of lists of strings, one list for each device, with
           each list holding each part of the device output.
        :type data: str or Iterable[str] or Iterable[Iterable[str]]
        :return: A list of parsed devices.
        :rtype: List[Device]
        """
        if isinstance(data, str):
            data = data.splitlines()
        return list(map(self.parse_line, data))

    def parse_line(self, args: Union[str, Iterable[str]]) -> Device:
        """
        Parse a single line from lspci -mm into a single device, either
        as the line or as a list of fields.

        :param args: Line or list of fields to parse from.
        :type args: str or Iterable[str]
        :return: A single parsed device.
        :rtype: Device
        """
        if isinstance(args, str):
            args = shlex.split(args)
        return Device(**vars(self._parser.parse_args(args)))

    def run(self, **kwargs: Any) -> List[Device]:
        if kwargs.get('verbose'):
            raise ValueError(
                'Verbose output is unsupported from the SimpleParser. '
                'Please use the pylspci.parsers.VerboseParser instead.'
            )
        return super().run(**kwargs)



import re
from abc import ABC, abstractmethod
from typing import Any, ClassVar, Dict, Optional, Pattern, Type, TypeVar

from pylspci.fields import hexstring

T = TypeVar('T', bound='Filter')


class Filter(ABC):

    _REGEX: ClassVar[Pattern]

    @classmethod
    def parse(cls: Type[T], value: str) -> T:
        if not value:
            return cls()
        match = cls._REGEX.match(value)
        data: Dict[str, str] = {}
        if match:
            data = {k: v for k, v in match.groupdict().items()
                    if v is not None}
        if not match or not data:
            raise ValueError('Value is not a valid filter string')
        return cls(**{
            k: hexstring(v)
            for k, v in data.items()
            if v != '' and v != '*'
        })

    @abstractmethod
    def __init__(self, **kwargs: Any) -> None:
        "Create a filter."


class SlotFilter(Filter):
    """
    Describes a slot filter, to filter devices geographically.

    Any field set to ``None`` will remove filtering.
    """

    domain: Optional[int] = None
    """
    Device domain, as a four-digit hexadecimal number.
    """

    bus: Optional[int] = None
    """
    Device bus, as a two-digit hexadecimal number.
    """

    device: Optional[int] = None
    """
    Device number, as a two-digit hexadecimal number, up to `0x1f`.
    """

    function: Optional[int] = None
    """
    The slot's function, as a single octal digit.
    """

    # [[domain:]bus:][device][.function]
    _REGEX: ClassVar[Pattern] = re.compile(
        r'^(?:(?:(?P<domain>(?:[0-9a-f]{1,4}|\*?)):)?'
        r'(?P<bus>(?:[0-9a-f]{1,2}|\*?)):)?'
        r'(?P<device>(?:[01]?[0-9a-f]|\*?))?'
        r'(?:\.(?P<function>(?:[0-7]|\*?)))?$'
    )

    def __init__(self, *,
                 domain: Optional[int] = None,
                 bus: Optional[int] = None,
                 device: Optional[int] = None,
                 function: Optional[int] = None,
                 ):
        self.domain = domain
        self.bus = bus
        self.device = device
        self.function = function

    def __repr__(self) -> str:
        return '{}(domain={!r}, bus={!r}, device={!r}, function={!r})'.format(
            self.__class__.__name__,
            self.domain, self.bus, self.device, self.function,
        )

    def __str__(self) -> str:
        return '{}:{}:{}.{}'.format(*map(
            lambda x: '{:x}'.format(x) if x is not None else '',
            (self.domain, self.bus, self.device, self.function),
        ))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (self.domain, self.bus, self.device, self.function) == \
            (other.domain, other.bus, other.device, other.function)


class DeviceFilter(Filter):
    """
    Describes a device filter, to filter devices logically.

    Any field set to ``None`` will remove filtering.
    """

    cls: Optional[int] = None
    """
    Device class ID, as a four-digit hexadecimal number.
    """

    vendor: Optional[int] = None
    """
    Device vendor ID, as a four-digit hexadecimal number.
    """

    device: Optional[int] = None
    """
    Device ID, as a four-digit hexadecimal number.
    """

    # [vendor]:[device][:class]
    _REGEX: ClassVar[Pattern] = re.compile(
        r'^(?P<vendor>(?:[0-9a-f]{1,4}|\*?)):'
        r'(?P<device>(?:[0-9a-f]{1,4}|\*?))'
        r'(?::(?P<cls>(?:[0-9a-f]{1,4}|\*?))?)?$'
    )

    def __init__(self, *,
                 cls: Optional[int] = None,
                 vendor: Optional[int] = None,
                 device: Optional[int] = None,
                 ):
        self.cls = cls
        self.vendor = vendor
        self.device = device

    def __repr__(self) -> str:
        return '{}(cls={!r}, vendor={!r}, device={!r})'.format(
            self.__class__.__name__, self.cls, self.vendor, self.device,
        )

    def __str__(self) -> str:
        return ':'.join(map(
            lambda x: '{:x}'.format(x) if x is not None else '',
            (self.vendor, self.device, self.cls),
        ))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (self.vendor, self.device, self.cls) == \
            (other.vendor, other.device, other.cls)



#!/usr/bin/env python3
from typing import List

from setuptools import find_packages, setup


def read_requirements(filename: str) -> List[str]:
    return [req.strip() for req in open(filename)]


requirements = read_requirements('requirements.txt')
dev_requirements = read_requirements('requirements-dev.txt')

setup(
    name='pylspci',
    version=open('VERSION').read().strip(),
    author='Lucidiot',
    packages=find_packages(
        exclude=["*.tests", "*.tests.*", "tests.*", "tests"],
    ),
    entry_points={
        'console_scripts': ['pylspci=pylspci.__main__:main'],
    },
    package_data={
        '': [
            'VERSION',
            'LICENSE',
            'README.rst',
            'requirements.txt',
            'requirements-dev.txt',
        ],
        'pylspci': ['py.typed'],
    },
    python_requires='>=3.5',
    install_requires=requirements,
    extras_require={
        'dev': dev_requirements,
    },
    test_suite='pylspci.tests',
    license='GNU General Public License 3',
    description="Simple parser for lspci -mmnn.",
    long_description=open('README.rst').read(),
    long_description_content_type='text/x-rst',
    keywords="lspci parser",
    url="https://gitlab.com/Lucidiot/pylspci",
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Natural Language :: English",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Software Development :: Libraries",
        "Topic :: System :: Hardware",
        "Topic :: Utilities",
        "Typing :: Typed",
    ],
    project_urls={
        "Source Code": "https://gitlab.com/Lucidiot/pylspci",
        "GitHub Mirror": "https://github.com/Lucidiot/pylspci",
    }
)
